# document.py

# noinspection PyUnresolvedReferences
from sockets_communication.process import *
# noinspection PyUnresolvedReferences
from sockets_communication.sockets import *
# noinspection PyUnresolvedReferences
from sockets_communication.exceptions import *
# noinspection PyUnresolvedReferences
from sockets_communication.service import *

from sockets_communication.base import document

class Documentation:
    """"""
# end Documentation

document(Documentation)